/* EXTANT.C - handy-dandy dynamically resizing (but still static) lists */
/* Copyright (c) 1999 TTK Ciar / Bill Moyer
 * All rights reserved.
 *
 * This code is derived from code developed by TTK Ciar and Bill Moyer.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 *  1. Redistributions of source code must retain the above copyright 
 *     notice, this list of conditions and the following disclaimer.
 *  2. The name of the author may not be used to endorse or promote 
 *     products derived from this software without specific prior 
 *     written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* Adding elements to an extant will cause the extant to automatically 
   grow bigger as needed.  The scalar-element extants (xit, xlt, xpt) 
   all double in size with every grow, but the struct extant (xst) only
   grows incrementally.  Note that scalar-element extants are 75% memory
   efficient which makes them ~2.5x as memory efficient as a doubly-
   linked list, and 1.5x as memory efficient as a singly-linked list.
   Keeping all of the data together improves locality of access and thus
   reduces page swapping and cache/TLB misses.
 */

#define  _EXTANT_C_
#include <stdlib.h>
#include <string.h>
#include "extant.h"

int xt_nerr = 0;
int xt_error = 0;
#define XTERROR(X)    { xt_nerr++; xt_error = (X); return (X); }
#define XTOK(X)       {            xt_error =  0;  return (X); }
#define XTERRORN(X)   { xt_nerr++; xt_error = (X); return NULL; }
#define XTNOK(X)      {            xt_error =  0;  return (X); }

/* xit - extant of integer */
int   xit_init  ( xit *x, int n_ ) {  /* initialize *xit with n slots */
  int n = n_;
  if ( x == NULL ) XTERROR(-1)
  if ( n <= 0    ) n = 1;
  x->n   = 0;
  x->max = n;
  x->d   = (int*) malloc ( sizeof(int) * n );
  if ( x->d == NULL ) XTERROR(-2)
  XTOK(0)
  }

int   xit_fetch ( xit *x, int n ) {  /* return the n'th int in *xit  */
  if ( x == NULL ) XTERROR(-1)
  if ( n >= x->n ) XTERROR(-3)
  XTOK( x->d[n] )
  }

int   xit_add   ( xit *x, int d ) {  /* append n to the end of *xit  */
  if ( x == NULL   ) XTERROR(-1)
  if ( x->max == 0 ) XTERROR(-4)
  if ( x->n >= x->max ) {
    int *p = x->d;
    int i,k;
    k = x->n;
    x->max <<= 1;
    x->d = (int*) malloc ( sizeof(int) * x->max );
    if ( x->d == NULL ) {
      x->d = p;
      x->max >>= 1;
      XTERROR(-2)
      }
    for ( i = 0; i < k; i++ ) x->d[i] = p[i];
    free ( p );
    }
  x->d[x->n++] = d;
  XTOK(0)
  }

int   xit_del   ( xit *x, int i ) {
  if ( x == NULL ) XTERROR(-1)
  if ( i < 0     ) XTERROR(-3)
  if ( i >= x->n ) XTERROR(-4)
  if ( x->n > 1 ) {
    int j = x->n-1;
    x->d[i] = x->d[j];
    }
  x->n--;
  XTOK(0)
  }

void  xit_nuke  ( xit *x ) {
  free ( x->d );
  x->n   = 0;
  x->max = 0;
  xt_error = 0;
  }

/* xct - extant of character */
/* additional semantic: xct->d is always kept null-terminated */
int   xct_init  ( xct *x, int n_ ) {  /* initialize *xct with n slots */
  int n = n_;
  if ( x == NULL ) XTERROR(-1)
  if ( n <= 0    ) n = 1;
  x->n   = 0;
  x->max = n;
  x->d   = (char*) malloc ( n + 1 );
  x->d[0] = 0;
  if ( x->d == NULL ) XTERROR(-2)
  XTOK(0)
  }

char xct_fetch ( xct *x, int n ) {  /* return the n'th char in *xct  */
  if ( x == NULL ) XTERROR(-1)
  if ( n >= x->n ) XTERROR(-3)
  XTOK( x->d[n] )
  }

int xct_add   ( xct *x, char d ) {  /* append d to the end of *xct  */
  if ( x == NULL   ) XTERROR(-1)
  if ( x->max == 0 ) XTERROR(-4)
  if ( x->n >= x->max ) {
    char *p = x->d;
    int i,k;
    k = x->n;
    x->max <<= 1;
    x->d = (char*) malloc ( x->max + 1 );
    if ( x->d == NULL ) {
      x->d = p;
      x->max >>= 1;
      XTERROR(-2)
      }
    for ( i = 0; i < k; i++ ) x->d[i] = p[i];
    free ( p );
    x->d[k] = 0;
    }
  x->d[x->n++] = d;
  x->d[x->n] = 0;
  XTOK(0)
  }

int xct_addstr ( xct *x, char *str ) {  /* append *str to the end of *xct  */
  int slen, i;
  if ( x   == NULL ) XTERROR(-1)
  if ( str == NULL ) XTERROR(-1)
  if ( x->max == 0 ) XTERROR(-4)
  slen = strlen ( str );
  if ( x->n + slen > x->max ) {
    char *p = x->d;
    int i,k;
    k = x->n;
    x->max += slen;
    x->d = (char*) malloc ( x->max + 1 );
    if ( x->d == NULL ) {
      x->d = p;
      x->max -= slen;
      XTERROR(-2)
      }
    for ( i = 0; i < k; i++ ) x->d[i] = p[i];
    free ( p );
    x->d[k] = 0;
    }
  for ( i = 0; i < slen; i++ ) 
    x->d[x->n + i] = str[i];
  x->n += slen;
  x->d[x->n] = 0;
  XTOK(0)
  }

/*  Append slen bytes from *str to the end of *xct.  This function is used 
    when str is not *really* a string, but rather an array (and hence might 
    contain null bytes which do not indicate the end of the data).  When str
    is really a string, use xct_addstrn instead.  */
int xct_addn ( xct *x, char *str, int slen ) {  
  int i;
  if ( x    == NULL ) XTERROR(-1)
  if ( str  == NULL ) XTERROR(-1)
  if ( slen == 0    ) XTOK(0)
  if ( slen <  0    ) XTERROR(-1)
  if ( x->max == 0  ) XTERROR(-4)
  if ( x->n + slen > x->max ) {
    char *p = x->d;
    int i,k;
    k = x->n;
    x->max += slen;
    x->d = (char*) malloc ( x->max + 1 );
    if ( x->d == NULL ) {
      x->d = p;
      x->max -= slen;
      XTERROR(-2)
      }
    for ( i = 0; i < k; i++ ) x->d[i] = p[i];
    free ( p );
    x->d[k] = 0;
    }
  for ( i = 0; ( i < slen ) && ( str[i] != 0 ); i++ ) 
    x->d[x->n + i] = str[i];
  x->n += i;
  x->d[x->n] = 0;
  XTOK(0)
  }

/*  append *str, up to slen bytes, to the end of *xct  */
int xct_addstrn ( xct *x, char *str, int slen_ ) {  
  int i;
  int slen;
  if ( x    == NULL ) XTERROR(-1)
  if ( str  == NULL ) XTERROR(-1)
  if ( slen == 0    ) XTOK(0)
  if ( slen <  0    ) XTERROR(-1)
  if ( x->max == 0  ) XTERROR(-4)
  slen = strlen ( str );
  if ( slen > slen_ ) slen = slen_;
  if ( x->n + slen > x->max ) {
    char *p = x->d;
    int i,k;
    k = x->n;
    x->max += slen;
    x->d = (char*) malloc ( x->max + 1 );
    if ( x->d == NULL ) {
      x->d = p;
      x->max -= slen;
      XTERROR(-2)
      }
    for ( i = 0; i < k; i++ ) x->d[i] = p[i];
    free ( p );
    x->d[k] = 0;
    }
  for ( i = 0; ( i < slen ) && ( str[i] != 0 ); i++ ) 
    x->d[x->n + i] = str[i];
  x->n += i;
  x->d[x->n] = 0;
  XTOK(0)
  }

int xct_del ( xct *x, int i ) {
  if ( x == NULL ) XTERROR(-1)
  if ( i < 0     ) XTERROR(-3)
  if ( i >= x->n ) XTERROR(-4)
  if ( x->n > 1 ) {
    int j = x->n-1;
    x->d[i] = x->d[j];
    }
  x->n--;
  XTOK(0)
  }

void xct_nuke ( xct *x ) {
  if ( x->d != NULL ) free ( x->d );
  x->n   = 0;
  x->max = 0;
  xt_error = 0;
  }

/* xlt - extant of long */
int   xlt_init  ( xlt *x, int n_ ) {
  int n = n_;
  if ( x == NULL ) XTERROR(-1)
  if ( n <= 0    ) n = 1;
  x->n   = 0;
  x->max = n;
  x->d   = (long*) malloc ( sizeof(long) * n );
  if ( x->d == NULL ) XTERROR(-2)
  XTOK(0);
  }

long  xlt_fetch ( xlt *x, int n ) {
  if ( x == NULL ) XTERROR(-1)
  if ( n >= x->n ) XTERROR(-3)
  XTOK( x->d[n] )
  }

int   xlt_add   ( xlt *x, long d ) {
  if ( x == NULL   ) XTERROR(-1)
  if ( x->max == 0 ) XTERROR(-4)
  if ( x->n >= x->max ) {
    long *p = x->d;
    int i,k;
    k = x->n;
    x->max <<= 1;
    x->d = (long*) malloc ( sizeof(long) * x->max );
    if ( x->d == NULL ) {
      x->d = p;
      x->max >>= 1;
      XTERROR(-2)
      }
    for ( i = 0; i < k; i++ ) x->d[i] = p[i];
    free ( p );
    }
  x->d[x->n++] = d;
  XTOK(0)
  }

int   xlt_del   ( xlt *x, int i ) {
  if ( x == NULL ) XTERROR(-1)
  if ( i < 0     ) XTERROR(-3)
  if ( i >= x->n ) XTERROR(-4)
  if ( x->n > 1 ) {
    int j = x->n-1;
    x->d[i] = x->d[j];
    }
  x->n--;
  XTOK(0)
  }

void  xlt_nuke  ( xlt *x ) {
  if ( x == NULL ) { xt_nerr++; xt_error = -1; return; }
  free ( x->d );
  x->n   = 0;
  x->max = 0;
  xt_error = 0;
  }


/* xpt - extant of pointer */
int   xpt_init  ( xpt *x, int n_ ) {
  int n = n_;
  if ( x == NULL ) XTERROR(-1)
  if ( n <= 0    ) n = 1;
  x->n   = 0;
  x->max = n;
  x->d   = (void**) malloc ( sizeof(void*) * n );
  if ( x->d == NULL ) XTERROR(-2)
  XTOK(0)
  }

void *xpt_fetch ( xpt *x, int n ) {
  if ( x == NULL ) XTERRORN(-1)
  if ( n >= x->n ) XTERRORN(-6)
  XTOK( x->d[n] )
  }

int   xpt_add   ( xpt *x, void *d ) {
  if ( x == NULL   ) XTERROR(-1)
  if ( x->max == 0 ) XTERROR(-4)
  if ( x->n >= x->max ) {
    void **p = x->d;
    int i,k;
    k = x->n;
    x->max <<= 1;
    x->d = (void**) malloc ( sizeof(void*) * x->max );
    if ( x->d == NULL ) {
      x->d = p;
      x->max >>= 1;
      XTERROR(-2)
      }
    for ( i = 0; i < k; i++ ) x->d[i] = p[i];
    free ( p );
    }
  x->d[x->n++] = d;
  XTOK(0)
  }

int   xpt_del   ( xpt *x, int i, nukef *func ) {
  void *p;
  if ( x == NULL ) XTERROR(-1)
  if ( i < 0     ) XTERROR(-3)
  if ( i >= x->n ) XTERROR(-4)
  p = x->d[i];
  if ( x->n > 1 ) {
    int   j = x->n-1;
    x->d[i] = x->d[j];
    }
  x->n--;
  if ( func != NULL ) func(p);
  XTOK(0)
  }

/* if nukef NULL, does not try to free() structs referenced by data elements */
void  xpt_nuke  ( xpt *x, nukef *func ) {  
  if ( x == NULL ) { xt_nerr++; xt_error = -1; return; }
  if ( x->d != NULL ) if ( func != NULL ) {
    int i, k;
    for ( i = 0, k = x->max; i < k; i++ ) 
      if ( func != NULL ) 
         func ( x->d[i] );
      else
         free ( x->d[i] );
    }
  free ( x->d );
  x->n   = 0;
  x->max = 0;
  xt_error = 0;
  }


/* xst - extant of struct */
int   xst_init   ( xst *x, int sz, int n_ ) {
  int n = n_;
  if ( x == NULL ) XTERROR(-1)
  if ( sz == 0   ) XTERROR(-5)
  if ( n  <= 0   ) n = 1;
  x->n    = 0;
  x->max  = n;
  x->sz   = sz;
  x->grow = n;
  x->d    = (char*) malloc ( sz * n );
  if ( x->d == NULL ) XTERROR(-2)
  XTOK(0)
  }

int   xst_fetch  ( xst *x, int n,  void* d ) {
  if ( x == NULL ) XTERROR(-1)
  if ( n >= x->n ) XTERROR(-3)
  memcpy ( (char*)d, &(((char*)(x->d))[n*x->sz]), x->sz );
  XTOK(0)
  }

void *xst_fetchp ( xst *x, int n ) {
  if ( x == NULL ) XTERRORN(-1)
  if ( n >= x->n ) XTERRORN(-6)
  XTNOK((void*)(&((char*)(x->d))[n*x->sz]))
  }

void *xst_fetchc ( xst *x, int n ) {
  char *retval;
  if ( x == NULL ) XTERRORN(-1)
  if ( n >= x->n ) XTERRORN(-6)
  if ( ( retval = (char*) malloc ( x->sz ) ) != NULL )
    memcpy ( retval, &(((char*)(x->d))[n*x->sz]), x->sz );
  else XTERRORN(-2)
  XTNOK(retval)
  }

int   xst_add    ( xst *x, void *d ) {
  int sz = x->sz;
  if ( x == NULL   ) XTERROR(-1)
  if ( x->max == 0 ) XTERROR(-1)
  if ( x->n >= x->max ) {
    char *p = (char*)(x->d);
    int   k;
    k = sz * x->n;
    x->max <<= 1;
    x->d = (void*) malloc ( sz * x->max );
    if ( x->d == NULL ) {
      x->d = p;
      x->max >>= 1;
      XTERROR(-2)
      }
    memcpy ( (char*)(x->d), p, k );
    free ( p );
    }
  memcpy ( &(((char*)(x->d))[(x->n++)*sz]), d, sz );
  XTOK(0)
  }

int   xst_del   ( xst *x, int i, nukef *func ) {
  void *p;
  if ( x == NULL ) XTERROR(-1)
  if ( i < 0     ) XTERROR(-3)
  if ( i >= x->n ) XTERROR(-4)
  p = &(((char*)(x->d))[i*x->sz]);
  if ( x->n > 1 ) {
    int   j = x->n-1;
    memcpy ( &(((char*)(x->d))[i]), &(((char*)(x->d))[i]), x->sz );
    }
  x->n--;
  if ( func == NULL ) free(p); else func(p);
  XTOK(0)
  }

void  xst_nuke   ( xst *x, nukef *func ) {
  if ( x == NULL ) { xt_nerr++; xt_error = -1; return; }
  if ( x->d != NULL ) if ( func != NULL ) {
    int i, k;
    for ( i = 0, k = x->max; i < k; i++ ) 
        func ( (void*)(&(((char*)(x->d))[i*x->sz])) );
    }
  free ( x->d );
  x->n   = 0;
  x->max = 0;
  xt_error = 0;
  }

