/* EXTANT.H - handy-dandy dynamically resizing (but still static) lists */
/* Copyright (c) 1999 TTK Ciar / Bill Moyer
 * All rights reserved.
 *
 * This code is derived from code developed by TTK Ciar and Bill Moyer.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 *  1. Redistributions of source code must retain the above copyright 
 *     notice, this list of conditions and the following disclaimer.
 *  2. The name of the author may not be used to endorse or promote 
 *     products derived from this software without specific prior 
 *     written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS 
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* Adding elements to an extant will cause the extant to automatically 
   grow bigger as needed.  The scalar-element extants (xit, xlt, xpt) 
   all double in size with every grow, but the struct extant (xst) only
   grows incrementally.  Note that scalar-element extants are 75% memory
   efficient which makes them ~2.5x as memory efficient as a doubly-
   linked list, and 1.5x as memory efficient as a singly-linked list.
   Keeping all of the data together improves locality of access and thus
   reduces page swapping and cache/TLB misses.
 */

#ifndef _EXTANT_H_
#define _EXTANT_H_

#ifndef _EXTANT_C_
extern int xt_nerr;
extern int xt_error;
#endif  

typedef struct xitag {
  int   n;
  int   max;
  int  *d;
  } xit;

typedef struct xctag {
  int   n;
  int   max;
  char *d;
  } xct;

typedef struct xltag {
  int   n;
  int   max;
  long *d;
  } xlt;

typedef struct xptag {
  int    n;
  int    max;
  void **d;
  } xpt;

typedef struct xstag {
  int   n;
  int   max;
  int   sz;
  int   grow;
  void *d;
  } xst;

typedef void nukef(void*);

/* xit - extant of integer */
int   xit_init  ( xit*, int );  /* initialize *xit with n slots */
int   xit_fetch ( xit*, int );  /* return the n'th int in *xit  */
int   xit_add   ( xit*, int );  /* append n to the end of *xit  */
int   xit_del   ( xit*, int );  /* delete n'th element */
void  xit_nuke  ( xit* );

/* xct - extant of char */
int   xct_init   ( xct*, int );
char  xct_fetch  ( xct*, int );
char *xct_fetchp ( xct*, int );
int   xct_add    ( xct*, char );
int   xct_addstr ( xct*, char* );
int   xct_addn   ( xct*, char*, int );
int   xct_addstrn( xct*, char*, int );
int   xct_del    ( xct*, int );  
void  xct_nuke   ( xct* );

/* xlt - extant of long */
int   xlt_init  ( xlt*, int );
long  xlt_fetch ( xlt*, int );
int   xlt_add   ( xlt*, long );
int   xlt_del   ( xlt*, int );  /* delete n'th element */
void  xlt_nuke  ( xlt* );

/* xpt - extant of pointer */
int   xpt_init  ( xpt*, int );
void *xpt_fetch ( xpt*, int );
int   xpt_add   ( xpt*, void* );
int   xpt_del   ( xpt*, int, nukef* );  /* delete n'th element */
void  xpt_nuke  ( xpt*, nukef* );  /* if nukef NULL, does not try to free() 
                                     structs referenced by data elements */

/* xst - extant of struct */
int   xst_init   ( xst*, int, int );   /* n1 = size of struct, n2 = #/slots */
int   xst_fetch  ( xst*, int, void* ); /* puts n'th struct into space provided */
void *xst_fetchp ( xst*, int );        /* returns pointer to n'th struct */
void *xst_fetchc ( xst*, int );        /* returns pointer to copy of n'th struct */
int   xst_add    ( xst*, void* );
int   xst_del    ( xst*, int, nukef* );  /* delete n'th element */
void  xst_nuke   ( xst*, nukef* );

#endif

